/*
 *  hash.c: Maintains hashtables.
 *
 *  Copyright (C) 2002 by the past and present ircd coders, and others.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  $Id: hash.c 100 2003-11-05 00:33:42Z stu $
 */

#include "stdinc.h"
#include "tools.h"
#include "user.h"
#include "channel.h"
#include "services.h"
#include "hash.h"
#include "cache.h"

/* Hash alghoritm based on ircu 2.10 implementation by Andrea Cocito "Nemesi" */

/* This hash function returns *exactly* N%HASHSIZE, where 'N'
 * is the string itself (included the trailing '\0') seen as 
 * a baseHASHSHIFT number whose "digits" are the bytes of the
 * number mapped through a "weight" transformation that gives
 * the same "weight" to caseless-equal chars, example:
 *
 * Hashing the string "Nick\0" the result will be:
 * N  i  c  k \0
 * |  |  |  |  `--->  ( (hash_weight('\0') * (HASHSHIFT**0) +
 * |  |  |  `------>    (hash_weight('k')  * (HASHSHIFT**1) +
 * |  |  `--------->    (hash_weight('c')  * (HASHSHIFT**2) +
 * |  `------------>    (hash_weight('i')  * (HASHSHIFT**3) +
 * `--------------->    (hash_weight('N')  * (HASHSHIFT**4)   ) % HASHSIZE
 *
 * It's actually a lot similar to a base transformation of the
 * text representation of an integer.
 * Looking at it this way seems slow and requiring unlimited integer
 * precision, but we actually do it with a *very* fast loop, using only 
 * short integer arithmetic and by means of two memory accesses and 
 * 3 additions per each byte processed.. and nothing else, as a side
 * note the distribution of real nicks over the hash table of this
 * function is about 3 times better than the previous one, and the
 * hash function itself is about 25% faster with a "normal" HASHSIZE
 * (it gets slower with larger ones and faster for smallest ones
 * because the hash table size affect the size of some maps and thus
 * the effectiveness of RAM caches while accesing them).
 */

/* The first prime bigger than the maximum character code (255) */
#define HASHSHIFT 257

#define HASHMAPSIZE (HASHSIZE + HASHSHIFT + 1)

/* Static memory structures */

/* We need a first function that, given an integer h between 0 and
 * HASHSIZE+HASHSHIFT, returns ( (h * HASHSHIFT) % HASHSIZE ) )
 * We'll map this function in this table
 */
static unsigned int hash_map[HASHMAPSIZE];

/* Then we need a second function that "maps" a char to its weitgh,
 * changed to a table this one too, with this macro we can use a char
 * as index and not care if it is signed or not, no.. this will not
 * cause an addition to take place at each access, trust me, the
 * optimizer takes it out of the actual code and passes "label+shift"
 * to the linker, and the linker does the addition :)
 */
static unsigned int hash_weight_table[CHAR_MAX - CHAR_MIN + 1];
#define hash_weight(ch) hash_weight_table[ch - CHAR_MIN]

/* The actual hash tables, both MUST be of the same HASHSIZE, variable
 * size tables could be supported but the rehash routine should also
 * rebuild the transformation maps, I kept the tables of equal size 
 * so that I can use one hash function and one transformation map
 */
static struct nickcache *nickTable[HASHSIZE];
static struct User      *clientTable[HASHSIZE];
static struct Channel   *channelTable[HASHSIZE];

/* init_hash()
 *
 * inputs       - NONE
 * output       - NONE
 * side effects - Initialize the maps used by hash
 *                functions and clear the tables
 */
void
init_hash(void)
{
  int i, weight_key;
  unsigned long l;
  unsigned long m;

  weight_key = random() % 256;  /* better than nothing --adx */

  /* Here is to what we "map" a char before working on it */
  for (i = CHAR_MIN; i <= CHAR_MAX; i++)
    hash_weight(i) = ((unsigned int) tolower(i)) ^ weight_key;

  /* Clear the hash tables first */
  for (l = 0; l < HASHSIZE; l++)
  {
    clientTable[l]      = NULL;
    channelTable[l]     = NULL;
    nickTable[i]        = NULL;
  }

  /* And this is our hash-loop "transformation" function, 
   * basically it will be hash_map[x] == ((x*HASHSHIFT)%HASHSIZE)
   * defined for 0<=x<=(HASHSIZE+HASHSHIFT)
   */
  for (m = 0; m < (unsigned long) HASHMAPSIZE; m++)
  {
    l = m;
    l *= (unsigned long) HASHSHIFT;
    l &= HASHSIZE - 1;
    hash_map[m] = (unsigned int) l;
  }
}

/* These are the actual hash functions, since they are static
 * and very short any decent compiler at a good optimization level
 * WILL inline these in the following functions
 */

static unsigned int
strhash(const char *n)
{
  unsigned int hash = 0;

  while (*n != '\0')
    hash = hash_map[hash + hash_weight(*n++)];

  return(hash);
}

/************************** Externally visible functions ********************/

/* Optimization note: in these functions I supposed that the CSE optimization
 * (Common Subexpression Elimination) does its work decently, this means that
 * I avoided introducing new variables to do the work myself and I did let
 * the optimizer play with more free registers, actual tests proved this
 * solution to be faster than doing things like tmp2=tmp->hnext... and then
 * use tmp2 myself which would have given less freedom to the optimizer.
 */

/* hash_add_client()
 *
 * inputs       - pointer to client
 * output       - NONE
 * side effects - Adds a client's name in the proper hash linked
 *                list, can't fail, user_p must have a non-null
 *                name or expect a coredump, the name is infact
 *                taken from user_p->nick
 */
void
hash_add_user(struct User *user_p)
{
  unsigned int hashv = strhash(user_p->nick);

  user_p->hnext = clientTable[hashv];
  clientTable[hashv] = user_p;
}

/* hash_add_channel()
 *
 * inputs       - pointer to channel
 * output       - NONE
 * side effects - Adds a channel's name in the proper hash linked
 *                list, can't fail. chptr must have a non-null name
 *                or expect a coredump. As before the name is taken
 *                from chptr->name, we do hash its entire lenght
 *                since this proved to be statistically faster
 */
void
hash_add_channel(struct Channel *chptr)
{
  unsigned int hashv = strhash(chptr->name);

  chptr->hnext = channelTable[hashv];
  channelTable[hashv] = chptr;
}

/* hash_del_client()
 *
 * inputs       - pointer to client
 * output       - NONE
 * side effects - Removes a Client's name from the hash linked list
 */
void
hash_del_client(struct User *user_p)
{
  unsigned int hashv = strhash(user_p->nick);
  struct User *tmp = clientTable[hashv];

  if (tmp != NULL)
  {
    if (tmp == user_p)
    {
      clientTable[hashv] = user_p->hnext;
      user_p->hnext = user_p;
    }
    else
    {
      while (tmp->hnext != user_p)
      {
        if ((tmp = tmp->hnext) == NULL)
          return;
      }

      tmp->hnext = tmp->hnext->hnext;
      user_p->hnext = user_p;
    }
  }
}

/* hash_del_channel()
 *
 * inputs       - pointer to client
 * output       - NONE
 * side effects - Removes the channel's name from the corresponding
 *                hash linked list
 */
void
hash_del_channel(struct Channel *chptr)
{
  unsigned int hashv = strhash(chptr->name);
  struct Channel *tmp = channelTable[hashv];

  if (tmp != NULL)
  {
    if (tmp == chptr)
    {
      channelTable[hashv] = chptr->hnext;
      chptr->hnext = chptr;
    }
    else
    {
      while (tmp->hnext != chptr)
      {
        if ((tmp = tmp->hnext) == NULL)
          return;
      }

      tmp->hnext = tmp->hnext->hnext;
      chptr->hnext = chptr;
    }
  }
}

/* find_client()
 *
 * inputs       - pointer to name
 * output       - NONE
 * side effects - New semantics: finds a client whose name is 'name'
 *                if can't find one returns NULL. If it finds one moves
 *                it to the top of the list and returns it.
 */
struct User *
find_user(const char *name)
{
  unsigned int hashv = strhash(name);
  struct User *user_p;

  if ((user_p = clientTable[hashv]) != NULL)
  {
    if (strcasecmp(name, user_p->nick))
    {
      struct User *prev;

      while (prev = user_p, (user_p = user_p->hnext) != NULL)
      {
        if (!strcasecmp(name, user_p->nick))
        {
          prev->hnext = user_p->hnext;
          user_p->hnext = clientTable[hashv];
          clientTable[hashv] = user_p;
          break;
        }
      }
    }
  }

  return(user_p);
}

/* hash_find_channel()
 *
 * inputs       - pointer to name
 * output       - NONE
 * side effects - New semantics: finds a channel whose name is 'name', 
 *                if can't find one returns NULL, if can find it moves
 *                it to the top of the list and returns it.
 */
struct Channel *
hash_find_channel(const char *name)
{
  unsigned int hashv = strhash(name);
  struct Channel *chptr;

  if ((chptr = channelTable[hashv]) != NULL)
  {
    if (strcasecmp(name, chptr->name))
    {
      struct Channel *prev;

      while (prev = chptr, (chptr = chptr->hnext) != NULL)
      {
        if (!strcasecmp(name, chptr->name))
        {
          prev->hnext = chptr->hnext;
          chptr->hnext = channelTable[hashv];
          channelTable[hashv] = chptr;
          break;
        }
      }
    }
  }

  return(chptr);
}
