
/*
 * Routines to maintain a list of online servers.
 *
 * Blitzed Services copyright (c) 2000-2002 Blitzed Services team
 *     E-mail: services@lists.blitzed.org
 * Based on ircservices-4.4.8:
 * copyright (c) 1996-1999 Andrew Church.
 *     E-mail: <achurch@dragonfire.net>
 * copyright (c) 1999-2000 Andrew Kempe.
 *     E-mail: <theshadow@shadowfire.org>
 *
 * This program is free but copyrighted software; see the file COPYING for
 * details.
 */

#include "stdinc.h"
#include "tools.h"
#include "user.h"
#include "services.h"
#include "send.h"
#include "memory.h"
#include "log.h"
#include "server.h"
#include "mystring.h"

/* $Id: servers.c 100 2003-11-05 00:33:42Z stu $ */

/*************************************************************************/

static dlink_list server_list;	/* Pointer to first server in the list */

/*
 * Points to the server last returned by findserver(). This should improve
 * performance during netbursts of NICKs.
 */

static void recursive_squit(struct Server *parent, const char *reason);

/* Return information on memory use. Assumes pointers are valid. */

void 
get_server_stats(long *nservers, long *memuse)
{
	struct Server *server;
    dlink_node *ptr;
	long mem;

	mem = sizeof(struct Server) * dlink_list_length(&server_list);
	DLINK_FOREACH(ptr, server_list.head)
    {
        server = ptr->data;
		mem += strlen(server->name) + 1;
    }   
	
    *nservers = dlink_list_length(&server_list);
	*memuse = mem;
}

#ifdef DEBUG_COMMANDS
void 
send_server_list(struct User *user)
{
	struct Server *server;
    dlink_node *ptr;

	notice(s_OperServ, user->nick, "Servers: %d", dlink_list_length(&server_list));
    DLINK_FOREACH(ptr, server_list.head)
    {
        server = ptr->data;
		notice(s_OperServ, user->nick, "%s [%s] {%s}", server->name,
		       server->hub ? server->hub->name : "no hub",
		       server->child ? server->child->name : "no child");
	}
}
#endif				/* DEBUG_COMMANDS */

/* Allocate a new Server structure, fill in basic values, link it to the
 * overall list, and return it. Always successful.
 */

static struct Server *
new_server(const char *servername)
{
	struct Server *server;

	server = scalloc(sizeof(struct Server), 1);
	server->name = sstrdup(servername);

    dlinkAdd(server, &server->node, &server_list);

	return server;
}

/* Remove and free a Server structure. */

static void 
delete_server(struct Server *server)
{
	if (debug >= 2)
		slog("debug: delete_server() called");

	free(server->name);
    dlinkDelete(&server->node, &server_list);
	free(server);

	if (debug >= 2)
		slog("debug: delete_server() done");
}

struct Server *
findserver(const char *servername)
{
	register struct Server *server;
    dlink_node *ptr;

	if (servername == NULL)
		return NULL;

    DLINK_FOREACH(ptr, server_list.head)
    {
        server = ptr->data;
        if(stricmp(servername, server->name) == 0)
            return server;
    }
	return NULL;
}

/* Handle a server SERVER command.
 * 	source = server's hub; !*source indicates this is our hub.
 * 	av[0]  = server's name
 */

void
do_server(const char *source, int ac, char **av)
{
	struct Server *server, *tmpserver;

	server = new_server(av[0]);
	server->t_join = time(NULL);
	server->child = NULL;
	server->sibling = NULL;

	if (!*source) 
    {
/*		send_sqlines();*/
		return;
	}

	server->hub = findserver(source);
	if (server->hub == NULL) 
    {
		/*
		 * This should NEVER EVER happen - but it's here while
		 * this function is being developed. Remove it someday.  
		 * That day is today!
		 */
		return;
	}

	if (server->hub->child == NULL) 
		server->hub->child = server;
    else 
    {
		tmpserver = server->hub->child;
		while (tmpserver->sibling)
			tmpserver = tmpserver->sibling;
		tmpserver->sibling = server;
	}
	return;
}

/* "SQUIT" all servers who are linked to us via the specified server by
 * deleting them from the server list. The parent server is not deleted,
 * so this must be done by the calling function.
 */

static void 
recursive_squit(struct Server *parent, const char *reason)
{
	struct Server *server, *nextserver;

	server = parent->child;
	if (debug >= 2)
		slog("recursive_squit, parent: %s", parent->name);
	while (server != NULL) 
    {
		if (server->child != NULL)
			recursive_squit(server, reason);
		if (debug >= 2)
			slog("recursive_squit, child: %s", server->name);
		nextserver = server->sibling;
		delete_server(server);
		server = nextserver;
	}
}

/* Handle a server SQUIT command.
 * 	av[0] = server's name
 * 	av[1] = quit message 
 */

void 
do_squit(const char *source, int ac, char **av)
{
	struct Server *server, *tmpserver;

	server = findserver(av[0]);

	if (server != NULL) 
    {
		recursive_squit(server, av[1]);
		if (server->hub) 
        {
			if (server->hub->child == server) 
				server->hub->child = server->sibling;
            else 
            {
				for (tmpserver = server->hub->child; tmpserver->sibling != NULL;
                        tmpserver = tmpserver->sibling) 
                {
					if (tmpserver->sibling == server) 
                    {
						tmpserver->sibling = server->sibling;
						break;
					}
				}
			}
			delete_server(server);
		}

	} 
    else 
    {
        /* FIXME: someone forgot to consider jupes.. duh.. */
		wallops(s_OperServ,
			"WARNING: Tried to quit non-existent server: \2%s",
			av[0]);
		slog("server: Tried to quit non-existent server: %s", av[0]);
		return;
	}
}
